import os

from .stress_detector import StressDetector, FEATURES
from .dataset import dataset
import numpy as np


def create_dataset(dataset_dir=None, dictionary_path=None):
    dataset(dataset_dir=dataset_dir, dictionary_path=dictionary_path, mode='collect_dataset')


def train_model(dataset_dir=os.getcwd(), features_path='data/complete_features.tsv', clf=None):
    sd = StressDetector(dataset_dir, FEATURES)
    sd.get_features(features_path)
    evaluation = sd.train(clf())
    print(f'F1 Score: {np.mean(evaluation["f1"])}')


def extend_dictionary(dataset_dir=None, dictionary_path=None, classifier='models/classifier_vot.pkl', scaler='models'
                                                                                                             '/scaler'
                                                                                                             '.pkl'):
    dataset(dataset_dir=dataset_dir, dictionary_path=dictionary_path, mode='extend_dictionary', model_path=classifier, scaler_path=scaler)


import math
import os
import re
import subprocess
import whisper
from tqdm import tqdm
from os import listdir
from pydub import AudioSegment
import wave
import requests
import xml.etree.ElementTree as Et
import csv
import textgrid
import rusyllab
from .stress_detector import StressDetector, FEATURES


class FileSizeError(Exception):
    pass


class NoDictProvided(Exception):
    pass


def clear_old_files(working_dir, directory='wav_tg_all'):
    if not os.path.exists(os.path.join(working_dir, directory)):
        pass
    else:
        files = os.listdir(os.path.join(working_dir, directory))
        for file in files:
            os.remove(os.path.join(os.path.join(working_dir, directory), file))


def get_files(working_dir):
    file_pairs = {}
    for file in listdir(working_dir):
        text = os.path.join(working_dir, file[:-4] + '.txt')
        wav = os.path.join(working_dir, file)
        if file.endswith('.wav') and os.path.exists(text):
            file_pairs.update({wav: text})
        elif file.endswith('.wav') and not os.path.exists(text):
            file_pairs.update({wav: None})
    return file_pairs


def preprocess_audio(input_files):
    for wav in input_files.keys():
        with wave.open(wav, "rb") as wave_file:
            frame_rate = wave_file.getframerate()
        sound = AudioSegment.from_file(wav, format='wav', frame_rate=frame_rate)
        sound.set_channels(1).set_frame_rate(16000).export(wav, format='wav')


def get_size(input_files):
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    for wav, text in input_files.items():
        file_stats = os.stat(wav)
        if file_stats.st_size == 0:
            return "0B"
        i = int(math.floor(math.log(file_stats.st_size, 1024)))
        p = math.pow(1024, i)
        s = round(file_stats.st_size / p, 2)
        with open(text, 'r', encoding='utf-8') as t:
            transcription = t.read().split()
        if s > 2 and size_name[i] == "GB":
            print(f'Your input file {wav} is too big for processing (>2GB).')
            raise FileSizeError
        if len(transcription) > 3000:
            print(f'Your input file {text} is too big for processing (>3000 words).')
            raise FileSizeError


def transcribe_audio(input_files):
    for wav in input_files.keys():
        if input_files[wav] is None:
            model = whisper.load_model("base")
            result = model.transcribe(wav)
            with open(wav[:-4] + ".txt", "w", encoding="utf-8") as o:
                o.write(result["text"].lower())


def preprocess_text(input_files):
    for text_file in input_files.values():
        with open(text_file, 'r', encoding='utf-8') as tr:
            text = tr.read()
        capitalized = {}
        for i in range(len(text.split())):
            if text.split()[i - 1][-1] not in '.?!…' and text.split()[i][0].isupper():
                capitalized.update({i: text.split()[i].strip('.?!…')})
        text_clear = []
        for i in range(len(text.split())):
            word_to_compare = text.split()[i].strip('.?!…').capitalize()
            if i in capitalized.keys() and word_to_compare == capitalized[i]:
                text_clear.append(word_to_compare)
            else:
                text_clear.append(text.split()[i].lower())
        with open(text_file, 'w', encoding='utf-8') as nw:
            nw.write(' '.join(text_clear))


def webmaus(input_files, out_path):
    url = "https://clarin.phonetik.uni-muenchen.de/BASWebServices/services/runPipeline"
    for wav, text in tqdm(input_files.items()):
        files = {
            'TEXT': open(text, 'rb'),
            'PIPE': (None, 'G2P_MAUS_PHO2SYL'),
            'LANGUAGE': (None, 'rus-RU'),
            'SIGNAL': open(wav, 'rb'),
            'OUTFORMAT': (None, 'TextGrid'),
        }
        res = requests.post(url, files=files)
        if res.status_code == 200:
            tree = Et.fromstring(res.text)
            res2 = requests.get(tree.find('downloadLink').text)
            open(out_path + os.path.basename(wav)[:-4] + '.TextGrid', 'wb').write(
                res2.content)


def build_dictionary(dictionary_path):
    with open(dictionary_path, 'r', encoding='utf-8') as f:
        paradigms = f.read().splitlines()
    stressed_dict = {}
    for paradigm in paradigms:
        if '#' in paradigm:
            paradigm = paradigm.replace('#', ',')
        for word in paradigm.split(','):
            stressed_dict.update({word.replace("'", "").replace("`", ""): word})
    return stressed_dict


def get_number_of_syllables(word):
    s = list(word)
    chunks = []
    for i in range(len(s)):
        if s[i] in 'уеэоаыяиёУЕЭОАЫЯИЁ' and i + 1 < len(s):
            chunks.append(s[i] + s[i + 1])
        elif s[i] in 'уеэоаыяиёУЕЭОАЫЯИЁ':
            chunks.append(s[i])
    stress_position = 0
    for i in range(len(chunks)):
        if "'" in chunks[i]:
            stress_position = i + 1
    if stress_position != 0:
        return stress_position


def collect_dataset(working_dir, input_files, dictionary, output_path, mode='collect_dataset'):
    n = 0
    all_data = []
    words_all = []
    words_dict = []
    words_new = []
    number_of_word_recordings_all = 0
    number_of_word_recordings_dict = 0
    number_of_word_recordings_new = 0
    for wav in input_files.keys():
        if os.path.exists(wav[:-4] + '.TextGrid'):
            transcription = textgrid.TextGrid.fromFile(wav[:-4] + '.TextGrid')
            for i in range(len(transcription[0])):
                word = transcription[0][i]
                if mode == 'collect_dataset':
                    if word.mark != "":
                        number_of_word_recordings_all += 1
                        if word.mark not in words_all:
                            words_all.append(word.mark)
                    if word.mark != "" and word.mark in dictionary.keys() and len(
                            transcription[2][i].mark.split('.')) > 1 \
                            and get_number_of_syllables(dictionary[word.mark]) is not None:
                        nr_syll = str(len(transcription[2][i].mark.split('.')))
                        stress_pos = str(len(transcription[2][i].mark.split('.')) - get_number_of_syllables(
                            dictionary[word.mark]) + 1)
                        if word.mark not in words_dict:
                            all_data.append([str(n), word.mark, stress_pos, nr_syll])
                            words_dict.append(word.mark)
                            n += 1
                        start_time = word.minTime * 1000
                        end_time = word.maxTime * 1000
                        sound = AudioSegment.from_file(wav, format='wav', frame_rate=16000)
                        extract = sound[start_time:end_time]
                        counter_wav = 1
                        output_filename = word.mark + '_' + str(counter_wav) + '.wav'
                        output_wav = os.path.join(os.path.join(working_dir, output_path), output_filename)
                        while os.path.exists(output_wav):
                            output_wav = output_wav.replace(f'_{counter_wav}.', f'_{counter_wav + 1}.')
                            counter_wav += 1
                        extract.export(output_wav, format='wav')
                        number_of_word_recordings_dict += 1
                elif mode == 'extend_dictionary':
                    if word.mark != "":
                        number_of_word_recordings_all += 1
                        if word.mark not in words_all:
                            words_all.append(word.mark)
                    if word.mark in dictionary.keys() and word.mark not in words_dict:
                        words_dict.append(word.mark)
                    elif word.mark != "" and word.mark not in ['обо', 'изо', 'подо',
                                                               'нибудь'] and word.mark not in dictionary.keys(
                    ) and re.search(r'у|е|э|о|а|ы|я|и|ё|У|Е|Э|О|А|Ы|Я|И|Ё', word.mark):
                        nr_syll = str(len(transcription[2][i].mark.split('.')))
                        stress_pos = 0
                        if word.mark not in words_new:
                            all_data.append([str(n), word.mark, stress_pos, nr_syll])
                            words_new.append(word.mark)
                            n += 1
                        start_time = word.minTime * 1000
                        end_time = word.maxTime * 1000
                        sound = AudioSegment.from_file(wav, format='wav', frame_rate=16000)
                        extract = sound[start_time:end_time]
                        counter_wav = 1
                        output_filename = word.mark + '_' + str(counter_wav) + '.wav'
                        output_wav = os.path.join(os.path.join(working_dir, output_path), output_filename)
                        while os.path.exists(output_wav):
                            output_wav = output_wav.replace(f'_{counter_wav}.', f'_{counter_wav + 1}.')
                            counter_wav += 1
                        extract.export(output_wav, format='wav')
                        number_of_word_recordings_new += 1
    with open('./data/training_words.tsv', 'w', newline='') as tw:
        tsv_output = csv.writer(tw, delimiter='\t')
        tsv_output.writerow(['', 'word', 'stress_pos', 'nr_syll'])
        for row in all_data:
            tsv_output.writerow(row)
    return number_of_word_recordings_all, number_of_word_recordings_dict, number_of_word_recordings_new, len(
        words_all), len(words_dict), len(words_new)


def get_words(working_dir, out_path):
    subprocess.call(["Rscript", "scripts/get_words.R", working_dir + '/', out_path + '/'])


def clear_files(input_path):
    for file in os.listdir(input_path):
        wav_file = file.strip('.TextGrid') + '.wav'
        if file.endswith('.TextGrid') and not os.path.exists(
                os.path.join(input_path, wav_file)):
            os.remove(os.path.join(input_path, file))


def dataset(dataset_dir, dictionary_path=None, mode=None, model_path=None, scaler_path=None):
    working_dir = os.path.dirname(dataset_dir)
    out_path = os.path.join(dataset_dir, 'wav_tg_all')
    if not os.path.exists(out_path):
        os.mkdir(out_path)
    clear_old_files(dataset_dir)
    print("Getting files and preprocessing audio and text...")
    files = get_files(dataset_dir)
    preprocess_audio(files)
    get_size(files)
    transcribe_audio(files)
    files = get_files(dataset_dir)
    preprocess_text(files)
    print("Creating TextGrid files...")
    # webmaus(files, working_dir)
    if dictionary_path is None:
        raise NoDictProvided('No dictionary provided.')
    dictionary = build_dictionary(dictionary_path)
    print("Collecting dataset...")
    if mode == 'collect_dataset':
        recordings_all, recordings_dict, recordings_new, words_all, words_dict, words_new = collect_dataset(working_dir,
                                                                                                            files,
                                                                                                            dictionary,
                                                                                                            out_path,
                                                                                                            mode='collect_dataset')
    elif mode == 'extend_dictionary':
        recordings_all, recordings_dict, recordings_new, words_all, words_dict, words_new = collect_dataset(working_dir,
                                                                                                            files,
                                                                                                            dictionary,
                                                                                                            out_path,
                                                                                                            mode='extend_dictionary')
    get_words(dataset_dir, out_path)
    clear_files(out_path)
    sd = StressDetector(out_path, FEATURES)
    if mode == 'collect_dataset':
        sd.preprocess()
        sd.get_features().to_csv('data/complete_features.tsv', sep='\t')
        with open('data/complete_features.tsv') as cf:
            number_of_syllables = len(cf.read().split('\n'))
        print(f"Dataset creation is finished. Out of {recordings_all} recordings of {words_all} words,"
              f" {recordings_dict} recordings of {words_dict} words were processed. {number_of_syllables - 2} syllables"
              f" were"
              f" extracted.")
    elif mode == 'extend_dictionary':
        sd.load_classifier(model_path, scaler_path)
        words_auto = 0
        words_to_add = []
        lines_to_add = []
        for file in tqdm(os.listdir(out_path)):
            if file.endswith('.wav'):
                word_ortho = os.path.basename(file).split('_')[0]
                syllables = rusyllab.split_words(f"{word_ortho}".split())
                if 'ё' in word_ortho and word_ortho not in words_to_add:
                    words_to_add.append(''.join(word_ortho.replace('ё', 'ё' + "'")))
                    words_auto += 1
                elif len(syllables) == 1:
                    for i in list(word_ortho):
                        if i in 'уеэоаыяиёУЕЭОАЫЯИЁ':
                            words_to_add.append(''.join(word_ortho.replace(i, i + "'")))
                else:
                    stresses = sd.classify(os.path.join(out_path, file), word_ortho)
                    word_for_dict = []
                    for syllable, result in zip(syllables, stresses):
                        if result == 1:
                            for i in list(syllable):
                                if i in 'уеэоаыяиёУЕЭОАЫЯИЁ':
                                    word_for_dict.append(''.join(syllable.replace(i, i + "'")))
                        else:
                            word_for_dict.append(syllable)
                    words_to_add.append(''.join(word_for_dict))
        for item in words_to_add:
            lines_to_add.append(item.replace("'", "") + '#' + ''.join(item))
        else:
            with open(dictionary_path, 'a', encoding='utf-8') as d:
                for item in sorted(set(lines_to_add)):
                    d.write(item + '\n')
        print(f"Done. Out of {words_all} words, {words_dict} of them were already in the dictionary. {words_new} words"
              f" were added, {words_all - words_dict - words_new} ignored (as not having primary "
              f"stress).")



#!/usr/bin/env python3
# Author: Vera Bernhard // minor changes by Polina Mashkovtseva
# Date: 01.06.2021 // 31.05.2023

import os
import subprocess
import json
import re

from praatio import tgio
from typing import Tuple

import pandas as pd


# Class to process a recording of an isolated word:
# - Detects syllable within a words
# - Extracts duration, pitch, loudness and spectral features per syllabe


# http://clarin.phonetik.uni-muenchen.de/BASWebServices/services/runMAUSGetInventar?LANGUAGE=rus-RU
LANG_SAMPA_VOWELS = ['o~', 'E~', 'e~', 'a~', 'u', 'o', 'i', 'e', 'a', '1', '@U', 'EI', 'aU', 'aI']


class Sound:
    def __init__(self, ortho_word: str, sound_file: str) -> None:
        self.ortho = ortho_word
        if not os.path.isfile(sound_file):
            raise FileNotFoundError(
                f"'{sound_file}' can be acessed, check path")
        else:
            self.sound = os.path.abspath(sound_file)
        self.basename = os.path.basename(sound_file).replace('.wav', '')  # e.g. деньги_1
        self.tg_wav_path = os.path.dirname(self.sound)
        self.nr_syll, self.stress_pos = self.get_stats()

        self.tg = None

        # Prosogram Output
        self.pros_data = None
        self.pros_speaker_data = None

        # Extracted from Webmaus .TextGrid
        self.syll_tier_list = []
        self.vowel_tier_list = []

        self.features = None

        self.time_stamps = {}
        self.stress_index = -1

    def get_stats(self) -> Tuple[int, int]:
        """ extract the position of stress and nr of syllable
            from training_words.tsv

        Raises:
            KeyError: if word does not exist in training_words.tsv

        Returns:
            Tuple[int, int]: number of syllables, position of stress (counted from right to left)
        """

        all_words = pd.read_csv(
            'data/training_words.tsv', sep='\t', header=0)
        row = all_words.loc[all_words['word'] == self.ortho]
        if row.empty:
            raise KeyError(
                f"No stats on '{self.ortho}' found, check spelling of word")
        else:
            return row['nr_syll'].values[0], row['stress_pos'].values[0]

    def preprocess(self) -> None:
        """ opens already existing syllable alignment .TextGrid which must be in same directory as .wav,
                    runs prosogram, improves pitch
        """

        self.run_prosogram()
        self._improve_pitch()
        self._extract_segm_tg()

    def read_in(self) -> None:
        """requires webmaus (syllabification) and prosogramm (prosodic features extraction) already to be ran,
            reads in existing textgrid and prosogram output taken from tg_wav_path"""
        self.tg = self._get_tg_file()
        self._read_prosogram_result()
        self._extract_segm_tg()

    def run_prosogram(self) -> None:
        """ runs Prosogram on .wav and .TextGrid file

            if Prosogram fails with 'Using external segmentation in tier segm'
            strategy 'Automatic: acoustic syllables' is used

        Raises:
            subprocess.CalledProcessError: if the prosogram fallback strategy does not work
        """

        # rename tiers of .TextGrid to fit Prosogram
        self.tg = self._get_tg_file()
        self._rename_segm_tier()

        try:
            subprocess.run(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                            self.sound, '0', '0'],
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.STDOUT,
                           check=True)

            output = subprocess.check_output(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                                              self.sound, '0', '0'])

            if '*** ERROR ***' in output.decode('utf-8'):
                try:
                    subprocess.run(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                                    self.sound, '0', '1'],
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.STDOUT,
                                   check=True)
                    print(f"PROSOGRAM:\tsuccessfully run on '{self.basename}'")
                except subprocess.CalledProcessError:
                    print(f"ERROR - PROSOGRAM: fails on '{self.basename}'")
                    pass
            else:
                print(f"PROSOGRAM:\tsuccessfully run on '{self.basename}'")

        except subprocess.CalledProcessError:
            try:
                subprocess.run(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                                self.sound, '0', '1'],
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.STDOUT,
                               check=True)
                print(f"PROSOGRAM:\tsuccessfully run on '{self.basename}'")
            except subprocess.CalledProcessError:
                print(f"ERROR - PROSOGRAM: fails on '{self.basename}'")

        finally:
            self._read_prosogram_result()

    def _get_tg_file(self) -> str:
        """ returns the path of the .TextGrid file assuming that it
            it is in same directory as .wav and has same name as .wav

        Returns:
            str: path of .TextGrid path
        """
        return os.path.join(
            self.tg_wav_path, self.basename + '.TextGrid'
        )

    def _rename_segm_tier(self) -> None:
        """ Changes names of TextGrid tiers to fit Prosogram
            using change_tiername.sh
        """
        subprocess.run(['bash', 'scripts/bash/change_tiername.sh',
                       self.tg])

    def _improve_pitch(self) -> None:
        """ Recalculates pitch and reruns Prosogram with new .Pitch file

        Raises:
            subprocess.CalledProcessError: if calculating new .Pitch file fails
            subprocess.CalledProcessError: if rerunning Prosogram with new .Pitch fails
        """
        try:
            subprocess.run(['praat', '--run', 'scripts/praat/stylize_pitch.praat',
                            os.path.dirname(self.sound) + '/', self.basename],
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.STDOUT,
                           check=True)

        except subprocess.CalledProcessError:
            print(f'ERROR - PITCH: stylize pitch fails with {self.basename}')

        try:

            subprocess.run(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                            self.sound, '1', '0'],
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.STDOUT,
                           check=True)

        except subprocess.CalledProcessError:
            try:
                subprocess.run(['praat', '--run', 'scripts/praat/run_prosogram.praat',
                                self.sound, '1', '1'],
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.STDOUT,
                               check=True)
            except subprocess.CalledProcessError:
                print(f'ERROR - PROSOGRAM: Recalculate pitch fails with {self.basename}')

    @ property
    def syll_nr_agreeing(self) -> bool:
        """ checks whether prosogram has found the same number of syllables as predicted

        Raises:
            AttributeError: if self.pros_data does not exist, i.e. Prosogram was not run yet or failed

        Returns:
            bool:
        """
        try:
            if self.pros_data.shape[0] == self.nr_syll:
                return True

            else:
                return False

        # in case there is no _data.txt and hence no pros_data.shape
        except AttributeError:
            raise AttributeError(
                f"{self.basename} has not attribute 'pros_data': \n Prosogram not run")

    def _read_prosogram_result(self) -> None:
        """ Reads in _data.txt and _profile_data.txt created by Prosogram

        Raises:
            FileNotFoundError: if no _data.txt exists i.e. Prosogram failed
        """
        try:
            df = pd.read_csv(
                os.path.join(self.tg_wav_path, self.basename + "_data.txt"), sep="\t")
            self.pros_data = df

            df = pd.read_csv(
                os.path.join(self.tg_wav_path, self.basename + "_profile_data.txt"), sep="\t")
            self.pros_speaker_data = df

        # in case that Prosogram has failed and hence there is no _data.txt
        except FileNotFoundError:
            print(f'\nignoring {self.basename} as Prosogram failed to process syllables')

    def _extract_segm_tg(self) -> None:
        """ Extracts segments from .TextGrid (Webmaus forced syllable alignment):
            - syllable intervals
            - vowel intervals
            --> will be needed for backup strategy where Prosogram fails to detect syllables
        """
        tg_open = tgio.openTextgrid(self.tg)
        for interval in tg_open.tierDict['syll'].entryList:
            if interval.label != '<p:>':
                self.syll_tier_list.append(interval)

        for interval in tg_open.tierDict['segm'].entryList:
            if interval.label != '<p:>' and interval.label in LANG_SAMPA_VOWELS:
                self.vowel_tier_list.append(interval)

    def _prepare_features(self) -> None:
        """ generate basic features per syllable
            - pos: position of syllable
            - pros:
                -1: if Prosogram did not represent syllable
                -[0-3]: index of row in Prosogram data where syllable is represented
            - stressed: 1: stressed, 0: unstressed
            - file: .wav filename, e.g. деньги_1.wav
            - vowel: vowel included in nucles, e.g. u
            - nr_syll: number of syllables
            - start: start time of syllable nucleus
            - end: end time of syllable nucleus
        """
        data = []
        # add whether syllable is stressed or not
        for syll in range(self.nr_syll):
            # check if current syllable is stressed
            if self.nr_syll-self.stress_pos == syll:
                self.stress_index = syll
                data.append(1)
            else:
                data.append(0)
        self.features = pd.DataFrame({'stressed': data})

        self.features['pos'] = self.features.index
        self.features['pros_repr'] = self.features.index
        self.features['pros'] = 1
        self.features['file'] = self.basename
        self.features['vowel'] = [
            vowel_tier.label for vowel_tier in self.vowel_tier_list]
        self.features['nr_syll'] = self.nr_syll

        if not self.syll_nr_agreeing:
            # check which syllable is represented in prosogram output,
            # -1: if no information in prosogram output
            # index: for which line in prosogram output represents syll
            for index, row in self.pros_data.iterrows():
                start_time_nucl = float(row['nucl_t1'])
                end_time_nucl = float(row['nucl_t2'])

                for i, interval in enumerate(self.syll_tier_list):
                    # syllable is detected by Prosogram
                    # --> take time stamps of nucleus from Prosogram
                    if interval.start <= start_time_nucl <= interval.end:
                        self.features.at[i, 'pros_repr'] = index
                        self.features.at[i, 'start'] = start_time_nucl
                        self.features.at[i, 'end'] = end_time_nucl

                    # syllable is not detected by Prosogram
                    # --> take time stamps of nucleus from Webmaus .TextGrid
                    else:
                        self.features.at[i, 'pros_repr'] = -1
                        self.features.at[i,
                                         'start'] = self.vowel_tier_list[i].start
                        self.features.at[i,
                                         'end'] = self.vowel_tier_list[i].end
                        self.features.at[i, 'pros'] = 0
        else:
            for index, row in self.pros_data.iterrows():
                start_time_nucl = float(row['nucl_t1'])
                end_time_nucl = float(row['nucl_t2'])
                self.features.at[index, 'start'] = start_time_nucl
                self.features.at[index, 'end'] = end_time_nucl

        # prepare spectral information
        subprocess.check_output(['praat', '--run', 'scripts/praat/get_intensity_bands.praat',
                                 self.basename+'.wav', self.tg_wav_path])

    def get_features(self) -> pd.DataFrame:
        """ Calculates absolute, normalised and contextual features of all syllables in word
            in the acoustic domains of:
            - Duration
            - Loudness
            - Pitch
            - Spectrum

        Returns:
            pd.DataFrame: returns all features of word as a DataFrame
        """
        self._prepare_features()

        # get absolute features
        for index, row in self.features.iterrows():
            self._get_dur_features(index, row)
            self._get_int_features(index, row)
            self._get_pitch_features(index, row)
            self._get_spect_features(index, row)

        # get normalized features
        self._norm_dur_feat()
        self._norm_int_feat()
        self._norm_pitch_feat()

        # get context-aware features
        for index, row in self.features.iterrows():
            # is not first syllable of word
            if index > 0:
                left_syl = self.features.iloc[index-1]
                self._cont_dur_feat(index, row, 'left', left_syl)
                self._cont_int_feat(index, row, 'left', left_syl)
                self._cont_spect_feat(index, row, 'left', left_syl)
                self._cont_pitch_feat(index, row, 'left', left_syl)

            # is first syllable of word -> has not left neighbour
            else:
                self._cont_dur_feat(index, row, 'left')
                self._cont_int_feat(index, row, 'left')
                self._cont_spect_feat(index, row, 'left')
                self._cont_pitch_feat(index, row, 'left')

            # is not last syllable of word
            if index < self.nr_syll-1:
                right_syl = self.features.iloc[index+1]
                self._cont_dur_feat(index, row, 'right', right_syl)
                self._cont_int_feat(index, row, 'right', right_syl)
                self._cont_spect_feat(index, row, 'right', right_syl)
                self._cont_pitch_feat(index, row, 'right', right_syl)
            # is last syllable of word -> has no right neighbour
            else:
                self._cont_dur_feat(index, row, 'right')
                self._cont_int_feat(index, row, 'right')
                self._cont_spect_feat(index, row, 'right')
                self._cont_pitch_feat(index, row, 'right')

        # remove helper files needed for spectral features
        try:
            os.remove(os.path.join(self.tg_wav_path,
                                   self.basename + '_filt_500.wav'))
            os.remove(os.path.join(self.tg_wav_path,
                                   self.basename + '_filt_1000.wav'))
            os.remove(os.path.join(self.tg_wav_path,
                                   self.basename + '_filt_2000.wav'))
        except FileNotFoundError:
            pass

        return self.features

    def _get_dur_features(self, index: int, row: pd.Series) -> None:
        """ Calculates absolute durational features per syllable:
            - nucl_dur: duration of nucleus
            - syll_dur: duration of syllable

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
        """
        self.features.at[index, 'nucl_dur'] = row['end'] - row['start']

        pros_index = row['pros_repr']
        # Prosogram has detected syllable
        if pros_index > -1:
            # find corresponding line in pros_data
            pros_row = self.pros_data.iloc[pros_index]
            syll_dur = pros_row['syll_dur']
        # Prosogram has not found syllable
        # take duration of syllable from webmaus .TextGrid
        else:
            syll_dur = self.syll_tier_list[index].end - \
                self.syll_tier_list[index].start

        self.features.at[index, 'syll_dur'] = syll_dur

    def _get_int_features(self, index: int, row: pd.Series) -> None:
        """Calculates absolute loudness features per syllable:
            - rms: root mean square amplitude
            - int peak: maximum intensity

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
        """

        pros_index = row['pros_repr']

        # root mean square energy
        rms = subprocess.check_output(
            ['praat', '--run', 'scripts/praat/get_loudness.praat', self.tg_wav_path,
             self.basename, str(row['start']), str(row['end']), '1', '0', '0'])
        rms = float(rms.decode('utf-8').strip(' Pascal\n'))
        self.features.at[index, 'rms'] = rms

        # from prosogram: intensity peak
        if pros_index > -1:
            int_peak = self.pros_data.iloc[pros_index]['int_peak']
            self.features.at[index, 'int_peak'] = int_peak

        # fallback: intensity peak
        else:
            int_peak = subprocess.check_output(
                ['praat', '--run', 'scripts/praat/get_loudness.praat', self.tg_wav_path,
                    self.basename, str(row['start']), str(row['end']), '0', '1', '0'])
            if int_peak.decode('utf-8').strip(' dB\n') == '--undefined--':
                int_peak = 0.0
            else:
                int_peak = float(int_peak.decode('utf-8').strip(' dB\n'))
            self.features.at[index, 'int_peak'] = int_peak

    def _get_spect_features(self, index: int, row: pd.Series) -> None:
        """ Calculates absolute spectral features per syllable:
            - spect_b1: mean intensity between 500-1000 Hz
            - spect_b2: mean intensity between 1000-2000 Hz
            - spect_b3: mean intensity between 2000-4000 Hz

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
        """
        syl_int = subprocess.check_output(
            ['praat', '--run', 'scripts/praat/get_spectral.praat', self.tg_wav_path,
             self.basename, str(row['start']), str(row['end'])])
        syl_int = [float(x) if x != '--undefined--' else 0.0 for x in syl_int.decode('utf-8').split(' dB\n')[:-1]]
        self.features.at[index, 'spect_b1'] = syl_int[0]
        self.features.at[index, 'spect_b2'] = syl_int[1]
        self.features.at[index, 'spect_b3'] = syl_int[2]

    def _get_pitch_features(self, index: int, row: pd.Series) -> None:
        """ Calculates pitch features per syllable:
            - trajectory: sum of absolute pitch interval of tonal segments in nucleus after stylization by Prosogram
            - f0_max: maximal f0 in Hz
            - f0_mean: mean f0 in Hz
            - f0_meanST: mean f0 in emitones
            - hipitch: highest pitch stylized by Prosogram in Hz
            - intersyllab: difference in semitones between end of previous nucleus and start of current one

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
        """
        pros_index = row['pros_repr']

        # Prosogram has detected syllable
        if pros_index > -1:
            trajectory = self.pros_data.iloc[pros_index]['trajectory']
            f0_max = self.pros_data.iloc[pros_index]['f0_max']
            f0_mean = self.pros_data.iloc[pros_index]['f0_mean']

            self.features.at[index, 'trajectory'] = trajectory
            self.features.at[index, 'f0_max'] = f0_max
            self.features.at[index, 'f0_mean'] = f0_mean
            self.features.at[index,
                             'intersyllab'] = self.pros_data.iloc[pros_index]['intersyllab']
            self.features.at[index,
                             'f0_meanST'] = self.pros_data.iloc[pros_index]['f0_meanST']
            self.features.at[index,
                             'f0_max_styl'] = self.pros_data.iloc[pros_index]['hipitch']

        # Prosogram has not found syllable
        else:
            self.features.at[index, 'trajectory'] = 0.0
            self.features.at[index, 'intersyllab'] = 0.0
            self.features.at[index, 'f0_max_styl'] = 0.0

            # Fallback
            f0_output = subprocess.check_output(
                ['praat', '--run', 'scripts/praat/get_pitch.praat', self.tg_wav_path,
                 self.basename, str(row['start']), str(row['end'])])
            f0_output = re.split('\n| Hz| semitones re 1 Hz',
                                 f0_output.decode('utf-8'))

            f0_max, f0_mean, f0_meanST = [x for x in f0_output if x != '']

            try:
                self.features.at[index, 'f0_max'] = float(f0_max)

            except ValueError:
                self.features.at[index, 'f0_max'] = 0.0

            try:
                self.features.at[index, 'f0_mean'] = float(f0_mean)
            except ValueError:
                self.features.at[index, 'f0_mean'] = 0.0

            try:
                self.features.at[index, 'f0_meanST'] = float(f0_meanST)
            except ValueError:
                self.features.at[index, 'f0_meanST'] = 0.0

    def _norm_int_feat(self) -> None:
        """ Normalises the loudness features with
            average peak intensity and root mean square energy
        """
        start_word = str(self.features.iloc[0]['start'])
        end_word = str(self.features.iloc[0]['end'])

        avg_int = subprocess.check_output(
            ['praat', '--run', 'scripts/praat/get_loudness.praat', self.tg_wav_path,
             self.basename, start_word, end_word, '0', '0', '1'])
        if avg_int.decode('utf-8').strip(' dB\n') == '--undefined--':
            avg_int = 0.0
        else:
            avg_int = float(avg_int.decode('utf-8').strip(' dB\n'))
        avg_rms = self.features['rms'].mean()

        self.features['rms_norm'] = self.features['rms']-avg_rms
        self.features['int_peak_norm'] = self.features['int_peak']-avg_int

    def _norm_dur_feat(self) -> None:
        """ Normalises the durational features with
            average nucleus resp. syllable duration and average intrinstic vowel duration
            calculated on all training data
        """
        avg_nucl_dur = self.features['nucl_dur'].mean()
        avg_syll_dur = self.features['syll_dur'].mean()

        # normalise for speech rate
        self.features['nucl_dur_norm'] = self.features['nucl_dur'] / avg_nucl_dur
        self.features['syll_dur_norm'] = self.features['syll_dur'] / avg_syll_dur

        try:
            with open('data/vowel_length.json') as f:
                vowel_length = json.load(f)
                # normalise for intrinsic vowel length
                for index, row in self.features.iterrows():
                    vowel_avg = vowel_length[row['vowel']]
                    if vowel_avg > 0:
                        self.features.at[index,
                                         'nucl_dur_vnorm'] = row['nucl_dur_norm'] / vowel_avg
                    else:
                        self.features.at[index,
                                         'nucl_dur_vnorm'] = row['nucl_dur_norm']
        except FileNotFoundError:
            for index, row in self.features.iterrows():
                self.features.at[index, 'nucl_dur_vnorm'] = 0.00

    def _norm_pitch_feat(self) -> None:
        """ Normalises pitch features with
            average pitch in Hz and Semitonse
        """
        mean_pitch = self.pros_speaker_data.iloc[0]['F0MeanHz']
        mean_pitch_st = self.pros_speaker_data.iloc[0]['F0MeanInST']

        # measures in Hz
        self.features['f0_max_norm'] = self.features['f0_max'] - mean_pitch
        self.features['f0_mean_norm'] = self.features['f0_mean'] - mean_pitch
        self.features['f0_max_styl_norm'] = self.features['f0_max_styl'] - mean_pitch

        # measures in Semitones
        self.features['f0_meanST_norm'] = self.features['f0_meanST'] - \
            mean_pitch_st

    def _cont_dur_feat(self, index: int, row: pd.Series, side: str, neighbour=None) -> None:
        """ Calculates contextual durational features of a syllable
            by taking difference to left or right neighbouring syllable

            if there is no neighbour as it is first or last syllable,
            average of all syllables of word is taking as neighbour value

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
            side (str): {'left'/'right'} is considered
            neighbour ([type], optional): feature vector of neighbouring syllable Defaults to None.
        """
        avg_nucl_norm = self.features['nucl_dur_norm'].mean()
        avg_nucl_vnorm = self.features['nucl_dur_vnorm'].mean()
        avg_syll_norm = self.features['syll_dur_norm'].mean()

        if isinstance(neighbour, pd.Series):
            self.features.at[index,
                             f'nucl_dur_{side}'] = row['nucl_dur_norm'] - neighbour['nucl_dur_norm']
            self.features.at[index,
                             f'nucl_dur_v_{side}'] = row['nucl_dur_vnorm'] - neighbour['nucl_dur_vnorm']
            self.features.at[index,
                             f'syll_dur_{side}'] = row['syll_dur_norm'] - neighbour['syll_dur_norm']

        else:
            self.features.at[index,
                             f'nucl_dur_{side}'] = row['nucl_dur_norm'] - avg_nucl_norm
            self.features.at[index,
                             f'nucl_dur_v_{side}'] = row['nucl_dur_vnorm'] - avg_nucl_vnorm
            self.features.at[index,
                             f'syll_dur_{side}'] = row['syll_dur_norm'] - avg_syll_norm

    def _cont_int_feat(self, index: int, row: pd.Series, side: str, neighbour=None) -> None:
        """ Calculates contextual loudness features of a syllable
            by taking difference to left or right neighbouring syllable

            if there is no neighbour as it is first or last syllable,
            average of all syllables of word is taking as neighbour value

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
            side (str): {'left'/'right'} is considered
            neighbour ([type], optional): feature vector of neighbouring syllable Defaults to None.
        """
        avg_int = self.features['int_peak_norm'].mean()
        avg_rms = self.features['rms_norm'].mean()

        if isinstance(neighbour, pd.Series):
            self.features.at[index,
                             f'int_peak_{side}'] = row['int_peak_norm'] - neighbour['int_peak_norm']

            self.features.at[index,
                             f'rms_{side}'] = row['rms_norm'] - neighbour['rms_norm']

        else:
            self.features.at[index,
                             f'int_peak_{side}'] = row['int_peak_norm'] - avg_int
            self.features.at[index,
                             f'rms_{side}'] = row['rms_norm'] - avg_rms

    def _cont_spect_feat(self, index: int, row: pd.Series, side: str, neighbour=None) -> None:
        """  Calculates contextual spectral features of a syllable
            by taking difference to left or right neighbouring syllable

            if there is no neighbour as it is first or last syllable,
            average of all syllables of word is taking as neighbour value

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
            side (str): {'left'/'right'} is considered
            neighbour ([type], optional): feature vector of neighbouring syllable Defaults to None.
        """
        avg_b1 = self.features['spect_b1'].mean()
        avg_b2 = self.features['spect_b2'].mean()
        avg_b3 = self.features['spect_b3'].mean()

        if isinstance(neighbour, pd.Series):
            self.features.at[index,
                             f'spect_b1_{side}'] = row['spect_b1'] - neighbour['spect_b1']
            self.features.at[index,
                             f'spect_b2_{side}'] = row['spect_b2'] - neighbour['spect_b2']
            self.features.at[index,
                             f'spect_b3_{side}'] = row['spect_b3'] - neighbour['spect_b3']

        else:
            self.features.at[index,
                             f'spect_b1_{side}'] = row['spect_b1'] - avg_b1
            self.features.at[index,
                             f'spect_b2_{side}'] = row['spect_b2'] - avg_b2
            self.features.at[index,
                             f'spect_b3_{side}'] = row['spect_b3'] - avg_b3

    def _cont_pitch_feat(self, index: int, row: pd.Series, side: str, neighbour=None) -> None:
        """  Calculates contextual pitch features of a syllable
            by taking difference to left or right neighbouring syllable

            if there is no neighbour as it is first or last syllable,
            average of all syllables of word is taking as neighbour value

        Args:
            index (int): position of current syllable in word
            row (pd.Series): feature vector of current syllable
            side (str): {'left'/'right'} is considered
            neighbour ([type], optional): feature vector of neighbouring syllable Defaults to None.
        """
        avg_f0_max = self.features['f0_max_norm'].mean()
        avg_f0_mean = self.features['f0_mean_norm'].mean()
        avg_f0_max_styl = self.features['f0_max_styl_norm'].mean()
        avg_f0_meanST = self.features['f0_meanST_norm'].mean()

        if isinstance(neighbour, pd.Series):
            self.features.at[index,
                             f'f0_max_{side}'] = row['f0_max_norm'] - neighbour['f0_max_norm']
            self.features.at[index,
                             f'f0_mean_{side}'] = row['f0_mean_norm'] - neighbour['f0_mean_norm']
            self.features.at[index,
                             f'f0_max_styl_{side}'] = row['f0_max_styl_norm'] - neighbour['f0_max_styl_norm']
            self.features.at[index,
                             f'f0_meanST_{side}'] = row['f0_meanST_norm'] - neighbour['f0_meanST_norm']
        else:
            self.features.at[index,
                             f'f0_max_{side}'] = row['f0_max_norm'] - avg_f0_max
            self.features.at[index,
                             f'f0_mean_{side}'] = row['f0_mean_norm'] - avg_f0_mean
            self.features.at[index,
                             f'f0_max_styl_{side}'] = row['f0_max_styl_norm'] - avg_f0_max_styl
            self.features.at[index,
                             f'f0_meanST_{side}'] = row['f0_meanST_norm'] - avg_f0_meanST



from setuptools import setup, find_packages

setup(
    name='ruspeechstress',
    version='0.12',
    packages=find_packages(),
    url="https://github.com/pmashkovtseva/ruspeechstres",
    license='MIT',
    author='Polina Mashkovtseva',
    author_email='p.mashkovtseva@gmail.com',
    description='package for extending word stress dictionaries based on speech analysis',
)



#!/usr/bin/env python3
# Author: Vera Bernhard // minor changes by Polina Mashkovtseva
# Date: 21.06.2021 // 25.05.2023

import os
import json

import joblib
import pandas as pd
import numpy as np
from pandas import DataFrame

from sklearn.metrics import accuracy_score, precision_recall_fscore_support, f1_score
from sklearn.model_selection import ShuffleSplit, GridSearchCV
from sklearn.preprocessing import StandardScaler

from typing import Tuple, Dict, List

from .utils import plot_confusion_matrix
from tqdm import tqdm
from .sound import Sound, LANG_SAMPA_VOWELS

# selected features after feature engineering
FEATURES = ['pos', 'nr_syll', 'nucl_dur', 'syll_dur', 'syll_dur_norm',  'nucl_dur_left', 'nucl_dur_right', 'rms',
            'int_peak', 'rms_norm', 'int_peak_norm', 'rms_left', 'rms_right', 'int_peak_left', 'int_peak_right',
            'spect_b1', 'spect_b2', 'spect_b3', 'spect_b1_left', 'spect_b2_left', 'spect_b3_left', 'spect_b1_right',
            'spect_b2_right', 'spect_b3_right', 'trajectory', 'f0_max', 'f0_mean', 'f0_meanST', 'f0_max_styl',
            'f0_max_norm', 'f0_mean_norm', 'f0_max_styl_norm', 'f0_meanST_norm', 'intersyllab', 'f0_max_left',
            'f0_max_right', 'f0_mean_left', 'f0_mean_right', 'f0_max_styl_left', 'f0_max_styl_right', 'f0_meanST_left',
            'f0_meanST_right']


# Class to:
# - extract prosodic features on all recordings single words within a directory
# - train and evaluate sklearn classifiers on the resulted data
# - classify a single recording with trained classifier

class StressDetector:

    def __init__(self, tg_wav_path: str, features: List) -> None:
        self.tg_wav_path = tg_wav_path
        self.sound_list = []
        self.features = features
        self.wav_files = self.get_files_dir(self.tg_wav_path, 'wav')
        self.vowels_length = {v: [] for v in LANG_SAMPA_VOWELS}
        self.features_list = features
        self.scaler = None
        self.classifier = None

    def preprocess(self) -> None:
        """ Runs following preprocessing steps on all .wav files in self.tg_wav_path
            opens already existing syllable alignment .TextGrid which must be in same directory as .wav,
            runs prosogram, improves pitch

        Attention: Expects all .wav to be named in form деньги_1.wav (word_index.wav)

        """
        bar = tqdm(self.wav_files)
        for f in bar:
            try:
                ortho = os.path.basename(f).split('_')[0]
                s = Sound(ortho, f)
                bar.set_description(f'Preprocessing {s.basename.ljust(30)}')
                s.preprocess()
                if s.pros_data is not None:
                    self.sound_list.append(s)
            except IndexError:
                print(f"SKIPPED: file name '{os.path.basename(f)} is not of form word_index.wav")

    def read_in(self) -> None:
        """
            requires webmaus (syllabification) and prosogramm (prosodic features extraction) already to be ran,
            reads in existing textgrid and prosogram output for all files in tg_wav_path
        """
        bar = tqdm(self.wav_files)
        for f in bar:
            try:
                ortho = os.path.basename(f).split('_')[0]
                s = Sound(ortho, f)
                bar.set_description(f'Reading in {s.basename.ljust(30)}')
                s.read_in()
                if s.pros_data is not None:
                    self.sound_list.append(s)

            except IndexError:
                print(
                    f"SKIPPED: file name '{os.path.basename(f)} is not of form word_index.wav")

    def load_classifier(self, classifier: str, scaler: str) -> None:
        """loads the pretrained model and classifier

        Args:
            classifier (str): path to model
            scaler (str): path to scaler
        """
        self.classifier = joblib.load(classifier)
        self.scaler = joblib.load(scaler)

    @staticmethod
    def get_files_dir(path: str, type: str) -> List:
        """returns of a list of all files of a certain type e.g. .wav in a given direcotry

        Args:
            path (str): directory
            type (str): type, e.g. .wav

        Returns:
            List: files of type
        """
        return [os.path.join(path, f) for f in os.listdir(path)
                if os.path.isfile(os.path.join(path, f)) and f.endswith(type)]

    def get_features(self, tsv_file: str = None) -> pd.DataFrame:
        """ Calculates all features 

        Args:
            tsv_file (str, optional): file with all features of all data. Defaults to None.

        Returns:
            pd.DataFrame: all features 
        """
        if not tsv_file:
            self.features = pd.DataFrame()
            bar = tqdm(self.sound_list)
            for sound in bar:
                bar.set_description(
                    f'Feature extraction {sound.basename.ljust(30)}')
                self.features = self.features.append(
                    sound.get_features(), ignore_index=True)
            self.features.fillna(0, inplace=True)
        else:
            self.features = pd.read_csv(tsv_file, sep='\t', header=0)
            print(f"Successfully read in all features from '{tsv_file}'")
        return self.features

    def get_vowel_length(self, file_name: str) -> None:
        """ Caculates average vowel length on entire training data
            which is then used for duration normalisation

            Attention: has to be run after get_features()
        """
        for sound in self.sound_list:
            for index, row in sound.features.iterrows():
                vowel = sound.vowel_tier_list[index].label
                self.vowels_length[vowel].append(row['nucl_dur_norm'])
        for vowel, dur_list in self.vowels_length.items():
            if dur_list:
                self.vowels_length[vowel] = np.mean(dur_list)
            else:
                self.vowels_length[vowel] = 0.0
        with open(file_name, 'w', encoding='utf-8') as outfile:
            json.dump(self.vowels_length, outfile)

    def agree_syllable_nr(self) -> int:
        """ Counts the number of files where Prosogram detected all syllables

        Returns:
            int: nr of files where Prosogram found all syllables
        """
        syll_agree_c = 0
        for s in self.sound_list:
            if s.syll_nr_agreeing:
                syll_agree_c += 1
            else:
                print(f'SYLL DISAGREE {s.basename}')
        return syll_agree_c

    def test_features(self, classifier, name: str, features: List, outfile: str = None) -> Dict:
        """ Trains a given classifier with one feature of feature list at a time
            and saves f1 scores for each model in .tsv format

        Args:
            classifier: sklearn classifier
            name (str): name of classifier
            features (List): list of features
            outfile (str): name of outilfe

        Returns:
            Dict: dict with f1 score per feature
        """
        avg_performance = pd.DataFrame({'feature': features})
        with tqdm(total=len(features)) as pbar:
            for feat in features:
                pbar.set_description(f'Training {name} with {feat}')
                performance_dict = self.train(classifier, [feat])
                i = avg_performance.index[avg_performance['feature']
                                          == feat]
                avg_performance.at[i, 'f1'] = np.mean(performance_dict['f1'])
                pbar.update()
        if outfile:
            avg_performance.to_csv(outfile, sep='\t')

        else:
            return avg_performance

    def test_feature_groups(self, classifier, name: str, feature_groups: Dict, outfile: str = None) -> DataFrame:
        """ Trains a given classifier with each feature group of feature_groups at a time
            and saves f1 scores for each model in .tsv format

        Args:
            classifier ([type]): sklearn classifier
            name (str): name of classifier
            feature_groups (Dict): list of feature groups
            outfile (str): name of outilfe

        Returns:
            Dict: dict with f1 score per feature group
        """
        avg_performance = pd.DataFrame({'feature': feature_groups.keys()})
        with tqdm(total=len(feature_groups)) as pbar:
            for feat_group, features in feature_groups.items():
                pbar.set_description(f'Training {name} with {feat_group}')
                performance_dict = self.train(classifier, features)
                i = avg_performance.index[avg_performance['feature']
                                          == feat_group]
                avg_performance.at[i, 'f1'] = np.mean(performance_dict['f1'])
                pbar.update()
        if outfile:
            avg_performance.to_csv(outfile, sep='\t')
        else:
            return avg_performance

    def test_classifiers(self, classifiers: List, names: List, outfile=None, predict_post=True) -> DataFrame:
        """ Trains a list of classifiers with the features saved within this class

        Args:
            classifiers (List): list of sklearn classifiers
            names (List): names of sklearn classifiers
            outfile ([type], optional): if given, writes average F1 score per classifier in fiel. Defaults to None.
            predict_post (bool, optional): if set true, uses postprocessing to predict. Defaults to True.

        Returns:
            [type]: [description]
        """
        avg_performance = pd.DataFrame({'classifier': names})

        for name, clf in zip(names, classifiers):
            print(f'Currently training {name}')
            if predict_post:
                performance_dict = self.train(clf)
            else:
                performance_dict = self.train(
                    clf, predict_post=False)
            i = avg_performance.index[avg_performance['classifier'] == name]
            avg_performance.at[i, 'f1'] = np.mean(performance_dict['f1'])
            avg_performance.at[i, 'accuracy'] = np.mean(
                performance_dict['accuracy'])
            avg_performance.at[i, 'precision_0'] = np.mean(
                performance_dict['precision'][0])
            avg_performance.at[i, 'precision_1'] = np.mean(
                performance_dict['precision'][1])
            avg_performance.at[i, 'recall_0'] = np.mean(
                performance_dict['recall'][0])
            avg_performance.at[i, 'recall_1'] = np.mean(
                performance_dict['recall'][1])

        if outfile:
            avg_performance.to_csv(outfile, sep='\t')

        else:
            return avg_performance

    def train(self, clf, features: List = None, predict_post=True, matrix=False) -> Dict:
        """ Trains a given classifier on the feature list and returns
            different performance metrics

        Args:
            features (List): list of features to train
            clf: skleanr classifier
            predict_post (bool, optional): if set true, uses postprocessing to predict. Defaults to True.
            matrix (bool, optional): if set true, produces a normalized confusion matrix. Defaults to False.

        Returns:
            Dict: list of performance metric (precision, recall, accuracy, f1) for each fold of cross-validation
        """
        if not features:
            features = self.features_list
        performance_dict = {'precision': [[], []], 'recall': [[], []], 'accuracy': [], 'f1': []}

        total_true = []
        total_pred = []

        for train, test, test_files in self._split_dataset():
            d_train = train[features]
            t_train = train['stressed']
            if len(features) > 1:
                scaler = StandardScaler().fit(d_train)
                d_train = scaler.transform(d_train)
            clf.fit(d_train, t_train)
            test_true = []
            test_pred = []
            test_pred_proba = []

            for file in test_files:
                # returns all syllables of a word
                word = self.features.loc[self.features['file'] == file]
                test_true += word['stressed'].tolist()
                total_true += word['stressed'].tolist()
                word = word[features]
                if len(features) > 1:
                    word = scaler.transform(word)
                if predict_post:
                    probs, classes, _ = self.predict_post(clf, word)
                else:
                    probs, classes = self.predict(clf, word)
                test_pred += classes
                total_pred += classes
                test_pred_proba += probs

            accuracy = accuracy_score(test_true, test_pred)
            precision, recall, _, _ = precision_recall_fscore_support(
                test_true, test_pred)
            f1 = f1_score(test_true, test_pred)

            performance_dict['accuracy'].append(accuracy)
            performance_dict['f1'].append(f1)
            performance_dict['precision'][0].append(precision[0])
            performance_dict['recall'][0].append(recall[0])
            performance_dict['precision'][1].append(precision[1])
            performance_dict['recall'][1].append(recall[1])

        if matrix:
            plot_confusion_matrix(total_pred, total_true, [
                'unstressed', 'stressed'], 'conf.png')

        return performance_dict

    def grid_search(self, pipelines: List, params: List[Dict]) -> List[Dict]:
        """Peforms gridsearch on given classifiers and params list

        Args:
            pipelines (List): list of sklearn pipelines/classifiers
            params (List[Dict]): list of dicts with different hyperparameters

        Returns:
            List[Dict]: list of best best hyperparameters
        """
        best_params = []
        for i in range(len(pipelines)):
            grid = GridSearchCV(
                estimator=pipelines[i],
                param_grid=params[i],
                n_jobs=-1,  # uses all possible processor
                scoring='f1',
                cv=10,  # cross-validation
                verbose=2
            ).fit(self.features[self.features_list], self.features['stressed'])

            best_params.append((grid.best_score_, grid.best_params_))

        return best_params

    def _split_dataset(self):
        """Splits the training data 10-fold cross validated into 25/75
        while it makes sure that syllables of the same word """
        files = pd.unique(self.features['file'])
        rs = ShuffleSplit(test_size=0.25, n_splits=10, random_state=42)
        for train_index, test_index in rs.split(files):
            train_files = files[np.array(train_index)]
            test_files = files[np.array(test_index)]
            train = self.features.loc[self.features['file'].isin(
                train_files)]
            test = self.features.loc[self.features['file'].isin(
                test_files)]
            yield train, test, test_files

    def train_all(self, classifier, name: str, save=False) -> None:
        """ trains classifier on all training data,
            saves resulting scaler and model within class

        Args:
            classifier: classifier
            name (str): name of classifier
            save (bool, optional): if given: pickels model. Defaults to False.
        """

        train = self.features[self.features_list]
        target = self.features['stressed']
        scaler = StandardScaler().fit(train)
        train_scaled = scaler.transform(train)
        print(f'Currently Training {name} on all data')
        clf = classifier.fit(train_scaled, target)

        self.scaler = scaler
        self.classifier = clf
        self.clf_name = name

        if save:
            joblib.dump(scaler, 'models/scaler.pkl')
            joblib.dump(clf, f'models/classifier_{name}.pkl')

    def predict_post(self, clf, word: np.array) -> Tuple[List, List]:
        """ Classifies each syllable in a word, taking into account that there
            can only be one stress per word

        Args:
            clf ([type]): classifier
            word (np.array): word represent with feature vectors

        Returns:
            Tuple[List, List]: array of classification
            probabilites and an array of classes
        """

        # get probability for stressed syllable only
        classes = clf.predict(word).tolist()
        probs = clf.predict_proba(word)[:, 1].tolist()
        nr_stressed_syl = classes.count(1)
        # more or less than 1 stressed syllable detected
        if nr_stressed_syl != 1:
            # no stressed syllable found
            # --> chose syllable with highest probability to be stressed
            if nr_stressed_syl == 0:
                pred_stress_pos = np.argmax(probs)
                classes[pred_stress_pos] = 1

            # too many stressed syllables
            # --> chose of stressed syllabe the one with highest probability to be stressed
            if nr_stressed_syl > 1:
                max_prob = 0
                pred_stress_pos = 0
                syll = 0
                for cla, prob in zip(classes, probs):
                    if cla == 1 and prob > max_prob:
                        max_prob = prob
                        pred_stress_pos = syll
                    syll += 1
                classes = [0] * len(probs)
                classes[pred_stress_pos] = 1
            return probs, classes, pred_stress_pos

        # exactly 1 stressed syllable detected
        else:
            return probs, classes, classes.index(1)

    def predict(self, clf, word: np.array) -> Tuple[List, List]:
        """ predicts stress pattern of a word without post_processing

        Args:
            clf: classifier
            word (np.array): word respresented as feature vectors

        Returns:
            Tuple[List, List]: (probilities for stressed, stress pattern)
        """
        probs = clf.predict_proba(
            word)[:, 1].tolist()
        classes = clf.predict(word).tolist()

        return probs, classes

    def classify(self, wav_file: str, word: str) -> List:
        """

        Args:
            wav_file (str): path to audio file
            word (str): orthographic transcription of word
            feedback (bool, optional): If given: prints feedback. Defaults to False.

        Returns:
            List: (predicted stress pattern)
        """
        sound = Sound(word, wav_file)
        sound.preprocess()
        all_features = sound.get_features()
        feat_scaled = self.scaler.transform(all_features[self.features_list])
        _, classes, pred_stress_pos = self.predict_post(
            self.classifier, feat_scaled)

        pred = classes

        return pred


def main():
    pass


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Author: Vera Bernhard
# Date: 01.06.2021

import itertools
import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import confusion_matrix


# Sources:
# -----------------------------------
# Conufsion Matrix:
# based on https://medium.com/analytics-vidhya/generation-of-a-concatenated-confusion-matrix-in-cross-validation-912485c4a972


def generate_confusion_matrix(cnf_matrix, classes, normalize=False):
    if normalize:
        cnf_matrix = cnf_matrix.astype(
            'float') / cnf_matrix.sum(axis=1)[:, np.newaxis]
        print("Normalized confusion matrix")
    else:
        print('Confusion matrix, without normalization')

    plt.imshow(cnf_matrix, interpolation='nearest',
               cmap=plt.get_cmap('Blues'))
    plt.colorbar()

    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=45)
    plt.yticks(tick_marks, classes)

    fmt = '.2f' if normalize else 'd'
    thresh = cnf_matrix.max() / 2.

    for i, j in itertools.product(range(cnf_matrix.shape[0]), range(cnf_matrix.shape[1])):
        plt.text(j, i, format(cnf_matrix[i, j], fmt), horizontalalignment="center",
                 color="white" if cnf_matrix[i, j] > thresh else "black")
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    plt.tight_layout()

    return cnf_matrix


def plot_confusion_matrix(predicted_labels_list, y_test_list, class_names, plot_name):
    cnf_matrix = confusion_matrix(y_test_list, predicted_labels_list)
    np.set_printoptions(precision=2)

    # Plot normalized confusion matrix
    plt.figure()
    generate_confusion_matrix(cnf_matrix, classes=class_names,
                              normalize=True)
    plt.savefig(plot_name)


def main():
    pass


if __name__ == "__main__":
    main()
